// 
// The LibSea Graph Library.
// Copyright (C) 2000,2001,2002 The Regents of the University of California.
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 
// ######END_HEADER######
// 


package org.caida.libsea;

class GrowableStringArray
{
    ////////////////////////////////////////////////////////////////////////
    // CONSTRUCTORS
    ////////////////////////////////////////////////////////////////////////

    public GrowableStringArray()
    {
	this(DEFAULT_SEGMENT_SIZE);
    }

    // {segmentSize} must be a power of 2.
    public GrowableStringArray(int segmentSize)
    {
	if (segmentSize < MIN_SEGMENT_SIZE)
	{
	    segmentSize = MIN_SEGMENT_SIZE;
	}

	if ((segmentSize & -segmentSize) != segmentSize)
	{
	    throw new IllegalArgumentException("segmentSize must be a power of 2");
        }

	m_segmentSize = segmentSize;
	m_segmentMask = segmentSize - 1;
	m_segmentSizeBits = MiscMath.log2(m_segmentSize);
    }

    ////////////////////////////////////////////////////////////////////////
    // PUBLIC METHODS
    ////////////////////////////////////////////////////////////////////////

    public boolean isEmpty()
    {
	return getNumAllocated() == 0;
    }

    public String getValue(int index)
    {
	if (index >= m_numAllocatedSlots)
	{
	    throw new UnallocatedElementException();
	}

	return getValueUnchecked(index);
    }

    public void setValue(int index, String value)
    {
	if (index >= m_numAllocatedSlots)
	{
	    throw new UnallocatedElementException();
	}

	setValueUnchecked(index, value);
    }

    public int append(String value)
    {
	if (m_numAllocatedSlots == m_numUsableSlots)
	{
	    if (m_numAllocatedSegments == m_segments.length)
	    {
		expandSegmentsArray();
	    }

	    m_segments[m_numAllocatedSegments++] = new String[m_segmentSize];
	    m_numUsableSlots += m_segmentSize;
	}

	int retval = m_numAllocatedSlots++;
	setValueUnchecked(retval, value);

	return retval;
    }

    public void clear()
    {
	m_numAllocatedSegments = 0;
	m_segments = new String[STARTING_NUM_SEGMENTS][];
	m_numUsableSlots = 0;
	m_numAllocatedSlots = 0;
    }

    public int getNumAllocated()
    {
	return m_numAllocatedSlots;
    }

    public int getNumFree()
    {
	return m_numAllocatedSegments * m_segmentSize - m_numAllocatedSlots;
    }

    ////////////////////////////////////////////////////////////////////////
    // PRIVATE METHODS
    ////////////////////////////////////////////////////////////////////////

    private String getValueUnchecked(int index)
    {
	int segment = index >> m_segmentSizeBits;
	int offset = index & m_segmentMask;

	return m_segments[segment][offset];
    }

    private void setValueUnchecked(int index, String value)
    {
	int segment = index >> m_segmentSizeBits;
	int offset = index & m_segmentMask;

	m_segments[segment][offset] = value;
    }

    private void expandSegmentsArray()
    {
	String[][] oldSegments = m_segments;
	m_segments = new String[oldSegments.length * 2][];
	System.arraycopy(oldSegments, 0, m_segments, 0, oldSegments.length);
    }

    ////////////////////////////////////////////////////////////////////////
    // PRIVATE CONSTANTS
    ////////////////////////////////////////////////////////////////////////

    private static final boolean DEBUG = true;

    private static final int STARTING_NUM_SEGMENTS = 512;
    private static final int DEFAULT_SEGMENT_SIZE = 1024;
    private static final int MIN_SEGMENT_SIZE = 128;

    ////////////////////////////////////////////////////////////////////////
    // PRIVATE FIELDS
    ////////////////////////////////////////////////////////////////////////

    private int  m_numAllocatedSegments = 0;
    private String[][]  m_segments = new String[STARTING_NUM_SEGMENTS][];
    private final int  m_segmentSize;
    private final int  m_segmentMask;
    private final int  m_segmentSizeBits;

    private int  m_numUsableSlots = 0;
    private int  m_numAllocatedSlots = 0;
}
