//
// The LibSea Graph Library.
// Copyright (C) 2000,2001,2002 The Regents of the University of California.
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 

// ASCII Graph File Format Lexer and Parser

header
{
package org.caida.libsea;
}

//==========================================================================
{
import java.util.List;
import java.util.ArrayList;
}
class GraphFileParser extends Parser;

options
{
   exportVocab = GraphFile;
   defaultErrorHandler = false;
}

tokens
{
   GRAPH_KEYWORD = "Graph";
   TRUE_KEYWORD = "T";
   FALSE_KEYWORD = "F";
   BOOL_KEYWORD = "bool";
   INT_KEYWORD = "int";
   FLOAT_KEYWORD = "float";
   DOUBLE_KEYWORD = "double";
   STRING_KEYWORD = "string";
   FLOAT3_KEYWORD = "float3";
   DOUBLE3_KEYWORD = "double3";
   ENUM_KEYWORD = "enum";
   LIST_KEYWORD = "list";
   INT;
   FLOAT;
   DOUBLE;
}

{
    public static void main(String[] args)
    {
        try
        {
            GraphFileLexer lexer =
                new GraphFileLexer(new java.io.DataInputStream(System.in));
            GraphFileParser parser = new GraphFileParser(lexer);
            parser.dumpTokens();
        }
        catch(Exception e)
        {
            System.err.println("ERROR: " + e);
        }
    }

    /////////////////////////////////////////////////////////////////////////

    private void addLink(Token token, int source, int destination)
        throws RecognitionException
    {
        int id = m_builder.addLink(source, destination);
        if (id != m_numLinksAdded)
        {
            String msg = makeCompleteLocationString(token)
                + "link ID[" + id + "] should have been " + m_numLinksAdded;
            throw new InternalErrorException(msg);
        }
        ++m_numLinksAdded;
    }

    /////////////////////////////////////////////////////////////////////////

    private void addPath(Token token)
        throws RecognitionException
    {
        int id = m_builder.addPath();
        if (id != m_numPathsAdded)
        {
            String msg = makeCompleteLocationString(token)
                + "path ID[" + id + "] should have been " + m_numPathsAdded;
            throw new InternalErrorException(msg);
        }
        ++m_numPathsAdded;
    }

    /////////////////////////////////////////////////////////////////////////

    private void addPathLink(Token token)
        throws RecognitionException, TokenStreamException
    {
        int link = parseNonnegativeInt(token, "path link");
        validateLinkID(token, link, "path link");
        m_builder.addPathLink(link);
    }

    /////////////////////////////////////////////////////////////////////////

    private EnumerationCreator addEnumeration(Token token, String name)
        throws RecognitionException
    {
        try
        {
	    EnumerationCreator retval = m_builder.addEnumeration(name);
            if (retval.getEnumerationID() != m_numEnumerationsAdded)
            {
                String msg = makeCompleteLocationString(token)
                    + "enumeration ID[" + retval.getEnumerationID()
                    + "] should have been " + m_numEnumerationsAdded;
                throw new InternalErrorException(msg);
            }
            ++m_numEnumerationsAdded;
            return retval;
        }
        catch (DuplicateObjectException e)
        {
            String msg = makeDuplicateObjectErrorString(token, "enumeration");
            throw new SemanticException(msg, getFilename(), token.getLine());
        }
    }

    /////////////////////////////////////////////////////////////////////////

    private void addEnumerator(Token token, String name, int value,
                               EnumerationCreator creator)
        throws RecognitionException
    {
        try
        {
	    int id = creator.addEnumerator(name, value);
            if (id != m_numEnumeratorsAdded)
            {
                String msg = makeCompleteLocationString(token)
                    + "enumerator ID[" + id + "] should have been "
	            + m_numEnumeratorsAdded;
                throw new InternalErrorException(msg);
            }
	    m_enumeratorOwner.add(new Integer(creator.getEnumerationID()));
            ++m_numEnumeratorsAdded;
        }
        catch (DuplicateObjectException e)
        {
            String msg = makeDuplicateObjectErrorString(token, "enumerator");
            throw new SemanticException(msg, getFilename(), token.getLine());
        }
    }

    /////////////////////////////////////////////////////////////////////////

    private void addAttributeDefinition
       (Token token, String name, ValueType type, int enumeration, String expr)
        throws RecognitionException
    {
        try
        {
            int id = m_builder.addAttributeDefinition
                (name, type, enumeration, expr);
            if (id != m_numAttributesAdded)
            {
                String msg = makeCompleteLocationString(token)
                    + "attribute definition ID[" + id
                    + "] should have been " + m_numAttributesAdded;
                throw new InternalErrorException(msg);
            }
            ++m_numAttributesAdded;
        }
        catch (DuplicateObjectException e)
        {
            String msg =
                makeDuplicateObjectErrorString(token, "attribute definition");
            throw new SemanticException(msg, getFilename(), token.getLine());
        }
    }

    /////////////////////////////////////////////////////////////////////////

    private QualifierCreator addQualifier
        (Token token, String type, String name, String description)
        throws RecognitionException
    {
        try
        {
            return m_builder.addQualifier(type, name, description);
        }
        catch (DuplicateObjectException e)
        {
            String msg = makeDuplicateObjectErrorString(token, "qualifier");
            throw new SemanticException(msg, getFilename(), token.getLine());
        }
    }

    /////////////////////////////////////////////////////////////////////////

    private void addQualifierAttribute
        (Token token, int attribute, String alias, QualifierCreator creator)
        throws RecognitionException
    {
        try
        {
	    creator.associateAttribute(attribute, alias);
        }
        catch (DuplicateObjectException e)
        {
            String msg =
                makeDuplicateObjectErrorString(token, "qualifier attribute");
            throw new SemanticException(msg, getFilename(), token.getLine());
        }
    }

    /////////////////////////////////////////////////////////////////////////

    private void addFilter(Token token, String name, String expr)
        throws RecognitionException
    {
        int id = m_builder.addFilter(name, expr);
        if (id != m_numFiltersAdded)
        {
            String msg = makeCompleteLocationString(token) + "filter ID["
                + id + "] should have been " + m_numFiltersAdded;
            throw new InternalErrorException(msg);
        }
        ++m_numFiltersAdded;
    }

    /////////////////////////////////////////////////////////////////////////

    private SelectorCreator addSelector(Token token, String name)
        throws RecognitionException
    {
        SelectorCreator retval = m_builder.addSelector(name);
        if (retval.getSelectorID() != m_numSelectorsAdded)
        {
            String msg = makeCompleteLocationString(token)
                + "selector ID[" + retval.getSelectorID()
                + "] should have been " + m_numSelectorsAdded;
            throw new InternalErrorException(msg);
        }
        ++m_numSelectorsAdded;
        return retval;
    }

    /////////////////////////////////////////////////////////////////////////

    private void addSelectorMapping
        (Token token, String characteristic, int filter,
         boolean node, boolean link, boolean path, SelectorCreator creator)
        throws RecognitionException
    {
        try
        {
            creator.addSelectorMapping
                (characteristic, filter, node, link, path); 
        }
        catch (DuplicateObjectException e)
        {
            String msg =
                makeDuplicateObjectErrorString(token, "selector mapping");
            throw new SemanticException(msg, getFilename(), token.getLine());
        }
    }

    /////////////////////////////////////////////////////////////////////////

    private DisplayCreator addDisplay(Token token, String name)
        throws RecognitionException
    {
        DisplayCreator retval = m_builder.addDisplay(name);
        if (retval.getDisplayID() != m_numDisplaysAdded)
        {
            String msg = makeCompleteLocationString(token)
                 + "display ID[" + retval.getDisplayID()
                 + "] should have been " + m_numDisplaysAdded;
            throw new InternalErrorException(msg);
        }
        ++m_numDisplaysAdded;
        return retval;
    }

    /////////////////////////////////////////////////////////////////////////

    private void addDisplayMapping
        (Token token, String characteristic, int attribute,
         boolean node, boolean link, boolean path, DisplayCreator creator)
        throws RecognitionException
    {
        try
        {
            creator.addDisplayMapping
                (characteristic, attribute, node, link, path); 
        }
        catch (DuplicateObjectException e)
        {
            String msg =
                makeDuplicateObjectErrorString(token, "display mapping");
            throw new SemanticException(msg, getFilename(), token.getLine());
        }
    }

    /////////////////////////////////////////////////////////////////////////

    private void addPresentation(Token token, String name,
                                 int display, int selector)
        throws RecognitionException
    {
        int id = m_builder.addPresentation(name, display, selector);
        if (id != m_numPresentationsAdded)
        {
            String msg = makeCompleteLocationString(token) + "presentation ID["
                + id + "] should have been " + m_numPresentationsAdded;
            throw new InternalErrorException(msg);
        }
        ++m_numPresentationsAdded;
    }

    /////////////////////////////////////////////////////////////////////////

    private void validateNodeID(Token token, int id, String description)
        throws SemanticException
    {
        if (id >= m_numNodes)
        {
            String msg = makePartialLocationString(token) + description
                + "[" + id + "] is outside the range of valid node IDs[<"
	        + m_numNodes + "]";
            throw new SemanticException(msg, getFilename(), token.getLine());
        }
    }

    /////////////////////////////////////////////////////////////////////////

    private void validateLinkID(Token token, int id, String description)
        throws SemanticException
    {
        if (id >= m_numLinks)
        {
            String msg = makePartialLocationString(token) + description
                + "[" + id + "] is outside the range of valid link IDs[<"
	        + m_numLinks + "]";
            throw new SemanticException(msg, getFilename(), token.getLine());
        }
    }

    /////////////////////////////////////////////////////////////////////////

    private void validatePathID(Token token, int id, String description)
        throws SemanticException
    {
        if (id >= m_numPaths)
        {
            String msg = makePartialLocationString(token) + description
                + "[" + id + "] is outside the range of valid path IDs[<"
	        + m_numPaths + "]";
            throw new SemanticException(msg, getFilename(), token.getLine());
        }
    }

    /////////////////////////////////////////////////////////////////////////

    private void validateEnumerationID(Token token, int id, String description)
        throws SemanticException
    {
        if (id >= m_numEnumerationsAdded)
        {
            String msg = makePartialLocationString(token) + description + "["
                + id + "] is outside the range of added enumeration IDs[<"
	        + m_numEnumerationsAdded + "]";
            throw new SemanticException(msg, getFilename(), token.getLine());
        }
    }

    /////////////////////////////////////////////////////////////////////////

    private void validateEnumeratorID(Token token, int id, String description)
        throws SemanticException
    {
        if (id >= m_numEnumeratorsAdded)
        {
            String msg = makePartialLocationString(token) + description + "["
                + id + "] is outside the range of added enumerator IDs[<"
	        + m_numEnumeratorsAdded + "]";
            throw new SemanticException(msg, getFilename(), token.getLine());
        }
    }

    /////////////////////////////////////////////////////////////////////////

    private void validateAttributeID(Token token, int id, String description)
        throws SemanticException
    {
        if (id >= m_numAttributesAdded)
        {
            String msg = makePartialLocationString(token) + description + "["
                + id + "] is outside the range of added attribute IDs[<"
	        + m_numAttributesAdded + "]";
            throw new SemanticException(msg, getFilename(), token.getLine());
        }
    }

    /////////////////////////////////////////////////////////////////////////

    private void validateFilterID(Token token, int id, String description)
        throws SemanticException
    {
        if (id >= m_numFiltersAdded)
        {
            String msg = makePartialLocationString(token) + description
                + "[" + id + "] is outside the range of added filter IDs[<"
	        + m_numFiltersAdded + "]";
            throw new SemanticException(msg, getFilename(), token.getLine());
        }
    }

    /////////////////////////////////////////////////////////////////////////

    private void validateSelectorID(Token token, int id, String description)
        throws SemanticException
    {
        if (id >= m_numSelectorsAdded)
        {
            String msg = makePartialLocationString(token) + description
                + "[" + id + "] is outside the range of added selector IDs[<"
	        + m_numSelectorsAdded + "]";
            throw new SemanticException(msg, getFilename(), token.getLine());
        }
    }

    /////////////////////////////////////////////////////////////////////////

    private void validateDisplayID(Token token, int id, String description)
        throws SemanticException
    {
        if (id >= m_numDisplaysAdded)
        {
            String msg = makePartialLocationString(token) + description
                + "[" + id + "] is outside the range of added display IDs[<"
	        + m_numDisplaysAdded + "]";
            throw new SemanticException(msg, getFilename(), token.getLine());
        }
    }

    /////////////////////////////////////////////////////////////////////////

    private void validatePresentationID
	(Token token, int id, String description)
        throws SemanticException
    {
        if (id >= m_numPresentationsAdded)
        {
            String msg = makePartialLocationString(token) + description + "["
                + id + "] is outside the range of added presentation IDs[<"
	        + m_numPresentationsAdded + "]";
            throw new SemanticException(msg, getFilename(), token.getLine());
        }
    }

    /////////////////////////////////////////////////////////////////////////

    private int parseNonnegativeInt(Token token, String description)
        throws SemanticException, TokenStreamException
    {
        int n = parseInt(token.getText());
        if (n < 0)
        {
            String msg = makePartialLocationString(token) + description
                + " must be nonnegative";
            throw new SemanticException(msg, getFilename(), token.getLine());
        }
        return n;
    }

    /////////////////////////////////////////////////////////////////////////

    private int parseInt(String s)
        throws TokenStreamException
    {
        try
        {
	    return Integer.parseInt(s);
        }
        catch (NumberFormatException e)
        {
	    String msg =
                "INTENAL ERROR: malformed integer literal \"" + s + "\"";
            throw new TokenStreamException(msg);
        }
    }

    /////////////////////////////////////////////////////////////////////////

    private float parseFloat(String s)
        throws TokenStreamException
    {
        try
        {
	    return Float.parseFloat(s);
        }
        catch (NumberFormatException e)
        {
	    String msg =
                "INTERNAL ERROR: malformed float literal \"" + s + "\"";
            throw new TokenStreamException(msg);
        }
    }

    /////////////////////////////////////////////////////////////////////////

    private double parseDouble(String s)
        throws TokenStreamException
    {
        try
        {
	    return Double.parseDouble(s);
        }
        catch (NumberFormatException e)
        {
	    String msg =
                "INTERNAL ERROR: malformed double literal \"" + s + "\"";
            throw new TokenStreamException(msg);
        }
    }

    /////////////////////////////////////////////////////////////////////////

    private String makeDuplicateObjectErrorString(Token token,
                                                  String description)
    {
        return makePartialLocationString(token)
            + "duplicate " + description + " name";
    }

    /////////////////////////////////////////////////////////////////////////

    private String makeCompleteLocationString(Token token)
    {
        return "line " + token.getLine()
            + ", column " + token.getColumn() + ": ";
    }

    /////////////////////////////////////////////////////////////////////////

    private String makePartialLocationString(Token token)
    {
        return "column " + token.getColumn() + ": ";
    }

    /////////////////////////////////////////////////////////////////////////

    private void initialize(GraphBuilder builder)
    {
        m_builder = builder;
        m_enumeratorOwner = new ArrayList();
        m_numEnumerationsAdded = 0;
        m_numEnumeratorsAdded = 0;
        m_numAttributesAdded = 0;
        m_numLinksAdded = 0;
        m_numPathsAdded = 0;
        m_numFiltersAdded = 0;	
        m_numSelectorsAdded = 0;
        m_numDisplaysAdded = 0;
        m_numPresentationsAdded = 0;
    }

    /////////////////////////////////////////////////////////////////////////

    private GraphBuilder  m_builder;
    private List  m_enumeratorOwner;

    private int  m_numNodes;
    private int  m_numLinks;
    private int  m_numPaths;
    private int  m_numPathLinks;

    private int  m_numEnumerationsAdded;
    private int  m_numEnumeratorsAdded;
    private int  m_numAttributesAdded;
    private int  m_numLinksAdded;
    private int  m_numPathsAdded;
    private int  m_numFiltersAdded;	
    private int  m_numSelectorsAdded;
    private int  m_numDisplaysAdded;
    private int  m_numPresentationsAdded;

    private MenuParser  m_presentationMenuParser =
        new MenuParser(new MenuIDValidator(MenuType.PRESENTATION),
                       new TopLevelMenuCreator(MenuType.PRESENTATION));

    private MenuParser  m_displayMenuParser =
        new MenuParser(new MenuIDValidator(MenuType.DISPLAY),
                       new TopLevelMenuCreator(MenuType.DISPLAY));

    private MenuParser  m_selectorMenuParser =
        new MenuParser(new MenuIDValidator(MenuType.SELECTOR),
                       new TopLevelMenuCreator(MenuType.SELECTOR));

    private MenuParser  m_filterMenuParser =
        new MenuParser(new MenuIDValidator(MenuType.FILTER),
                       new TopLevelMenuCreator(MenuType.FILTER));

    private MenuParser  m_attributeMenuParser =
        new MenuParser(new MenuIDValidator(MenuType.ATTRIBUTE),
                       new TopLevelMenuCreator(MenuType.ATTRIBUTE));

    ///////////////////////////////////////////////////////////////////////
    // PRIVATE INNER CLASSES
    ///////////////////////////////////////////////////////////////////////

    private class TypeHolder
    {
        public ValueType type;
        public int enumeration;
    }

    /////////////////////////////////////////////////////////////////////////

    private class Attribute
        implements ValueConsumer
    {
        public Attribute(ValueType type, int enumeration)
        {
            m_valueType = type;
            m_enumeration = enumeration;
        }

	////////////////////////////////////////////////////////////////////
	// PUBLIC METHODS
	////////////////////////////////////////////////////////////////////

        public void targetNodes()
        {
            m_objectType = ObjectType.NODE;
        }

        public void targetLinks()
        {
            m_objectType = ObjectType.LINK;
        }

        public void targetPaths()
        {
            m_objectType = ObjectType.PATH;
        }

        public ValueConsumer addValue(Token token, int id)
            throws RecognitionException
        {
            reset();
            validateObjectID(token, id);
            m_creator = m_builder.addObjectAttribute(m_objectType, id);
            return this;
        }

        ////////////////////////////////////////////////////////////////////
        // PUBLIC METHODS (interface ValueConsumer)
        ////////////////////////////////////////////////////////////////////

        public void addBooleanValue(Token token, boolean value)
            throws RecognitionException
        {
            validateValueType(token, ValueType.BOOLEAN);
	    m_creator.addBooleanValue(value);
            ++m_numValuesAdded;
        }

        public void addIntegerValue(Token token, int value)
            throws RecognitionException
        {
            validateValueType(token, ValueType.INTEGER);
	    m_creator.addIntegerValue(value);
            ++m_numValuesAdded;
        }

        public void addFloatValue(Token token, float value)
            throws RecognitionException
        {
            validateValueType(token, ValueType.FLOAT);
	    m_creator.addFloatValue(value);
            ++m_numValuesAdded;
        }

        public void addDoubleValue(Token token, double value)
            throws RecognitionException
        {
            validateValueType(token, ValueType.DOUBLE);
	    m_creator.addDoubleValue(value);
            ++m_numValuesAdded;
        }

        public void addStringValue(Token token, String value)
            throws RecognitionException
        {
            validateValueType(token, ValueType.STRING);
	    m_creator.addStringValue(value);
            ++m_numValuesAdded;
        }

        public void addFloat3Value(Token token, float x, float y, float z)
            throws RecognitionException
        {
            validateValueType(token, ValueType.FLOAT3);
	    m_creator.addFloat3Value(x, y, z);
            ++m_numValuesAdded;
        }

        public void addDouble3Value(Token token, double x, double y, double z)
            throws RecognitionException
        {
            validateValueType(token, ValueType.DOUBLE3);
	    m_creator.addDouble3Value(x, y, z);
            ++m_numValuesAdded;
        }

	public void addEnumerationValue(Token token, int enumerator)
            throws RecognitionException
        {
            validateValueType(token, ValueType.ENUMERATION);
            validateEnumeratorOwnership(token, enumerator);
	    m_creator.addEnumerationValue(enumerator);
            ++m_numValuesAdded;
        }

	////////////////////////////////////////////////////////////////////
	// PRIVATE METHODS
	////////////////////////////////////////////////////////////////////

        private void validateValueType(Token token, ValueType seenType)
            throws SemanticException
        {
            if (seenType.getBaseType() != m_valueType.getBaseType())
            {
                String msg = makePartialLocationString(token)
                    + "wrong type of value for attribute: need "
                    + mapValueTypeToName(m_valueType) + " value";
                throw new SemanticException
                    (msg, getFilename(), token.getLine());
            }

            if (m_numValuesAdded > 0 && m_valueType.isBaseType())
            {
                String msg = makePartialLocationString(token)
                    + "wrong type of value for attribute: "
                    + "supplied a list of values for a scalar attribute";
                throw new SemanticException
                    (msg, getFilename(), token.getLine());
            }
        }

        private String mapValueTypeToName(ValueType type)
        {
            switch (m_valueType.getBaseType())
            {
            case ValueType._BOOLEAN: return "boolean";
            case ValueType._INTEGER: return "integer";
            case ValueType._FLOAT: return "float";
            case ValueType._DOUBLE: return "double";
            case ValueType._STRING: return "string";
            case ValueType._FLOAT3: return "float3";
            case ValueType._DOUBLE3: return "double3";
            case ValueType._ENUMERATION: return "enumeration";
            default: throw new InternalErrorException();
            }
        }

        private void validateObjectID(Token token, int id)
            throws SemanticException
        {
            switch (m_objectType.getType())
            {
            case ObjectType._NODE: validateNodeID(token, id, "node"); break;
            case ObjectType._LINK: validateLinkID(token, id, "link"); break;
            case ObjectType._PATH: validatePathID(token, id, "path"); break;
            default: throw new InternalErrorException();
            }
        }

        private void validateEnumeratorOwnership(Token token, int id)
            throws SemanticException
        {
            Integer integer = (Integer)m_enumeratorOwner.get(id);
            if (integer == null)
            {
                String msg = makeCompleteLocationString(token) +
                    "no ownership info found for " + "enumerator[" + id + "]";
                throw new InternalErrorException(msg);
            }
            int owner = integer.intValue();
            if (owner != m_enumeration)
            {
                String msg = makePartialLocationString(token) +
                    "enumerator[" + id + "] does not belong "
                    + "to the enumeration[" + m_enumeration
                    + "] of the current attribute";
                throw new SemanticException
                    (msg, getFilename(), token.getLine());
            }
        }

        private void reset()
        {
            m_creator = null;
            m_numValuesAdded = 0;
        }

	////////////////////////////////////////////////////////////////////
	// PRIVATE FIELDS
	////////////////////////////////////////////////////////////////////

        private ValueType  m_valueType;
        private int  m_enumeration;

        private ObjectType  m_objectType;

        private AttributeCreator  m_creator;
        private int  m_numValuesAdded;
    }

    /////////////////////////////////////////////////////////////////////////

    private class MenuIDValidator
        implements MenuType
    {
        public MenuIDValidator(int type)
        {
            m_type = type;
        }

        public void validateID(Token token, int id)
            throws RecognitionException
        {
            if (id >= 0)
            {
                switch (m_type)
                {
                case PRESENTATION:
                    validatePresentationID
                        (token, id, "presentation menu item");
                    break;
    
                case DISPLAY:
                    validateDisplayID(token, id, "display menu item");
                    break;
    
                case SELECTOR:
                    validateSelectorID(token, id, "selector menu item");
                    break;
    
                case FILTER:
                    validateFilterID(token, id, "filter menu item");
                    break;
    
                case ATTRIBUTE:
                    validateAttributeID(token, id, "attribute menu item");
                    break;
    
                default:
                    {
                        String msg = "invalid m_type[" + m_type + "]";
                        throw new InternalErrorException(msg);
                    }
                }
            }    
        }

        private int  m_type;
    }

    /////////////////////////////////////////////////////////////////////////

    private class TopLevelMenuCreator
        implements MenuType, MenuCreator
    {
        public TopLevelMenuCreator(int type)
        {
            m_type = type;
        }

        public MenuCreator addSubmenu(String label, int id)
        {
            switch (m_type)
            {
            case PRESENTATION:
                return m_builder.addPresentationSubmenu(label, id);

            case DISPLAY:
                return m_builder.addDisplaySubmenu(label, id);

            case SELECTOR:
                return m_builder.addSelectorSubmenu(label, id);

            case FILTER:
                return m_builder.addFilterSubmenu(label, id);

            case ATTRIBUTE:
                return m_builder.addAttributeSubmenu(label, id);

            default:
                {
                    String msg = "invalid m_type[" + m_type + "]";
                    throw new InternalErrorException(msg);
                }
            }
        }

        private int  m_type;
    }

    /////////////////////////////////////////////////////////////////////////

    private class MenuParser
    {
        public MenuParser(MenuIDValidator validator, MenuCreator creator)
        {
            m_validator = validator;
            m_creator = creator;
        }

        public MenuParser addSubmenu(Token token, String label, int id)
            throws RecognitionException
        {
            m_validator.validateID(token, id);
            MenuCreator submenuCreator = m_creator.addSubmenu(label, id);
            return new MenuParser(m_validator, submenuCreator);
        }

        private MenuIDValidator  m_validator;
        private MenuCreator  m_creator;
    }

    /////////////////////////////////////////////////////////////////////////

    private interface ValueConsumer
    {
        void addBooleanValue(Token token, boolean value)
            throws RecognitionException;
        void addIntegerValue(Token token, int value)
            throws RecognitionException;
        void addFloatValue(Token token, float value)
            throws RecognitionException;
        void addDoubleValue(Token token, double value)
            throws RecognitionException;
        void addStringValue(Token token, String value)
            throws RecognitionException;
        void addFloat3Value(Token token, float x, float y, float z)
            throws RecognitionException;
        void addDouble3Value(Token token, double x, double y, double z)
            throws RecognitionException;
        void addEnumerationValue(Token token, int enumerator)
            throws RecognitionException;
    }

    /////////////////////////////////////////////////////////////////////////

    private interface MenuType
    {
        int  PRESENTATION = 0;
        int  DISPLAY = 1;
        int  SELECTOR = 2;
        int  FILTER = 3;
        int  ATTRIBUTE = 4;
    }
}

file[GraphBuilder builder]
{
    initialize(builder);
}
	: g:graph EOF
	;

graph
	: GRAPH_KEYWORD  LCURLY
	  ( name:STRING
            { m_builder.setGraphName(name.getText()); }
          )?  SEMI
	  ( desc:STRING 
	   { m_builder.setGraphDescription(desc.getText()); }
          )?  SEMI
          nn:INT  SEMI
           { m_numNodes = parseNonnegativeInt(nn, "# nodes");
             m_builder.allocateNodes(m_numNodes); }
          nl:INT  SEMI
           { m_numLinks = parseNonnegativeInt(nl, "# links");
             m_builder.allocateLinks(m_numLinks); }
          np:INT  SEMI
           { m_numPaths = parseNonnegativeInt(np, "# paths");
             m_builder.allocatePaths(m_numPaths); }
          npl:INT  SEMI
           { m_numPathLinks = parseNonnegativeInt(npl, "# path links");
             m_builder.allocatePathLinks(m_numPathLinks); }
	  ( linkList )?  SEMI
          ( pathList )?  SEMI
          ( enumDefList )?  SEMI
	  ( attrDefList )?  SEMI
          ( qualifierList )?  SEMI
          ( filterList )?  SEMI
          ( selectorList )?  SEMI
          ( displayList )?  SEMI
          ( presentationList )?  SEMI
          ( menuList[m_presentationMenuParser] )?  SEMI
 	  ( menuList[m_displayMenuParser] )?  SEMI
          ( menuList[m_selectorMenuParser] )?  SEMI
    	  ( menuList[m_filterMenuParser] )?  SEMI
          ( menuList[m_attributeMenuParser] )?  SEMI
          RCURLY
	;

linkList
	: LBRACKET  ( link ( COMMA link )* )?  RBRACKET
	;

link
	: start:LCURLY  src:INT  SEMI  dest:INT  SEMI  RCURLY
          {
              int source = parseNonnegativeInt(src, "link source");
              int destination = parseNonnegativeInt(dest, "link destination");
              validateNodeID(src, source, "link source");
              validateNodeID(dest, destination, "link destination");
              addLink(start, source, destination);
          }
	;

pathList
	: LBRACKET  ( path ( COMMA path )* )?  RBRACKET
	;

path
	: start:LCURLY { addPath(start); } ( pathLinkList )?  SEMI  RCURLY
	;

pathLinkList
	: LBRACKET  
          ( first:INT { addPathLink(first); }
            ( COMMA rest:INT { addPathLink(rest); }
            )*
          )?  RBRACKET
	;

enumDefList
	: LBRACKET  ( enumDef ( COMMA enumDef )* )?  RBRACKET
	;

enumDef
{
    EnumerationCreator creator = null;
}
	: start:LCURLY  name:IDENT  SEMI
           { creator = addEnumeration(start, name.getText()); }
          ( enumeratorList[creator] )?  SEMI  RCURLY
	;

enumeratorList[EnumerationCreator creator]
	: LBRACKET
          ( enumerator[creator] ( COMMA enumerator[creator] )* )?
          RBRACKET
	;

enumerator[EnumerationCreator creator]
	: start:LCURLY  name:IDENT  SEMI  value:INT  SEMI  RCURLY
           { addEnumerator(start, name.getText(),
                           parseInt(value.getText()), creator); }
	;

attrDefList
	: LBRACKET  ( attrDef ( COMMA attrDef )* )?  RBRACKET
	;

attrDef
{
    TypeHolder holder = null;
    String expr = null;
    Attribute attribute = null;
}
	: start:LCURLY
          n:IDENT  SEMI  holder=t:attrType  SEMI  ( expr=attrDefValue )?  SEMI
          {
              String name = n.getText();
              ValueType type = holder.type;
              int enumeration = holder.enumeration;
              addAttributeDefinition(start, name, type, enumeration, expr);
              attribute = new Attribute(holder.type, holder.enumeration);
          }
          { attribute.targetNodes(); }
          ( objectAttrList[attribute] )?  SEMI
          { attribute.targetLinks(); }
          ( objectAttrList[attribute] )?  SEMI
          { attribute.targetPaths(); }
          ( objectAttrList[attribute] )?  SEMI
          RCURLY
	;

attrType returns [TypeHolder holder]
{
    boolean isList = false;
}
	: ( LIST_KEYWORD { isList = true; } )?  holder=baseType[isList]

	;

baseType[boolean isList] returns [TypeHolder holder]
{
    holder = new TypeHolder();
    holder.enumeration = -1;
}
	: ( BOOL_KEYWORD
            { holder.type =
                  (isList ? ValueType.BOOLEAN_LIST : ValueType.BOOLEAN ); }
          | INT_KEYWORD
            { holder.type =
                  (isList ? ValueType.INTEGER_LIST : ValueType.INTEGER ); }
          | FLOAT_KEYWORD
            { holder.type =
                  (isList ? ValueType.FLOAT_LIST : ValueType.FLOAT ); }
	  | DOUBLE_KEYWORD
            { holder.type =
                  (isList ? ValueType.DOUBLE_LIST : ValueType.DOUBLE ); }
          | STRING_KEYWORD
            { holder.type =
                  (isList ? ValueType.STRING_LIST : ValueType.STRING ); }
          | FLOAT3_KEYWORD
            { holder.type =
                  (isList ? ValueType.FLOAT3_LIST : ValueType.FLOAT3 ); }
	  | DOUBLE3_KEYWORD
            { holder.type =
                  (isList ? ValueType.DOUBLE3_LIST : ValueType.DOUBLE3 ); }
          | ENUM_KEYWORD  n:INT
            {
                int id = parseNonnegativeInt(n, "enumeration ID");
                validateEnumerationID(n, id, "enumeration type");
                holder.enumeration = id;
                holder.type = (isList ? ValueType.ENUMERATION_LIST
                                      : ValueType.ENUMERATION );
            }
          )
	;

attrDefValue returns [String expr]
	: c:CODE  { expr = c.getText(); }
	;

objectAttrList[Attribute attribute]
	: LBRACKET
          ( objectAttr[attribute] ( COMMA objectAttr[attribute] )* )?
          RBRACKET
	;

objectAttr[Attribute attribute]
{
    ValueConsumer consumer = null;
}
	: LCURLY  n:INT  SEMI
          {
              int id = parseNonnegativeInt(n, "object ID");
              consumer = attribute.addValue(n, id);
          }
          value[consumer]  SEMI  RCURLY
	;

qualifierList
	: LBRACKET  ( qualifier ( COMMA qualifier )* )?  RBRACKET
	;

qualifier
{ 
    QualifierCreator creator = null;
    String description = null;
}
	: start:LCURLY  type:IDENT  SEMI  name:IDENT  SEMI
          ( desc:STRING { description = desc.getText(); } )?  SEMI
          { creator = addQualifier
                (start, type.getText(), name.getText(), description); }
	  ( qualifierAttrList[creator] )?  SEMI  RCURLY
	;

qualifierAttrList[QualifierCreator creator]
	: LBRACKET  ( qualifierAttr[creator]
                      ( COMMA qualifierAttr[creator] )* )?  RBRACKET
	;

qualifierAttr[QualifierCreator creator]
	: start:LCURLY  n:INT  SEMI  id:IDENT  SEMI  RCURLY
          {
              int attribute = parseNonnegativeInt(n, "qualifier attribute");
              validateAttributeID(n, attribute, "qualifier attribute");
              String alias = id.getText();
              addQualifierAttribute(start, attribute, alias, creator);
          }
	;

filterList
	: LBRACKET  ( filter ( COMMA filter )* )?  RBRACKET
	;

filter
	: start:LCURLY  name:STRING  SEMI  expr:CODE  SEMI  RCURLY
          { addFilter(start, name.getText(), expr.getText()); }
	;

selectorList
	: LBRACKET  ( selector ( COMMA selector )* )?  RBRACKET
	;

selector
{
    SelectorCreator creator = null;
}
	: start:LCURLY  n:STRING  SEMI
          { creator = addSelector(start, n.getText()); }
          ( selectorMappingList[creator] )?  SEMI  RCURLY
	;

selectorMappingList[SelectorCreator creator]
	: LBRACKET  ( selectorMapping[creator]
                      ( COMMA selectorMapping[creator] )* )?  RBRACKET
	;

selectorMapping[SelectorCreator creator]
{
    boolean node = false;
    boolean link = false;
    boolean path = false;
}
	: start:LCURLY  f:INT  SEMI  c:STRING  SEMI
	  node=booleanValue  SEMI  link=booleanValue  SEMI
          path=booleanValue  SEMI  RCURLY
          {
              int filter = parseNonnegativeInt(f, "filter ID");
              validateFilterID(f, filter, "selector filter");
              addSelectorMapping(start, c.getText(), filter,
                                 node, link, path, creator);
          }
	;

displayList
	: LBRACKET  ( display ( COMMA display )* )?  RBRACKET
	;

display
{
    DisplayCreator creator = null;
}
	: start:LCURLY  n:STRING  SEMI
          { creator = addDisplay(start, n.getText()); }
          ( displayMappingList[creator] )?  SEMI  RCURLY
	;

displayMappingList[DisplayCreator creator]
	: LBRACKET  ( displayMapping[creator]
                      ( COMMA displayMapping[creator] )* )?  RBRACKET
	;

displayMapping[DisplayCreator creator]
{
    boolean node = false;
    boolean link = false;
    boolean path = false;
}
	: start:LCURLY  a:INT  SEMI  c:STRING  SEMI
	  node=booleanValue  SEMI  link=booleanValue  SEMI
          path=booleanValue  SEMI  RCURLY
          {
              int attribute = parseNonnegativeInt(a, "attribute ID");
              validateAttributeID(a, attribute, "display attribute");
              addDisplayMapping(start, c.getText(), attribute,
                                node, link, path, creator);
          }
	;

presentationList
	: LBRACKET  ( presentation ( COMMA presentation )* )?  RBRACKET
	;

presentation
{
    int selector = -1;
}
	: start:LCURLY  n:STRING  SEMI  d:INT  SEMI
          ( s:INT
           {
               selector = parseNonnegativeInt(s, "selector ID");
               validateSelectorID(s, selector, "presentation selector");
           }
          )?  SEMI  RCURLY
          {
               int display = parseNonnegativeInt(d, "display ID");
               validateDisplayID(d, display, "presentation display");
               addPresentation(start, n.getText(), display, selector);
          }
	;

menuList[MenuParser parser]
	: LBRACKET  ( menu[parser] ( COMMA menu[parser] )* )?  RBRACKET
	;

menu[MenuParser parser]
{
    MenuParser submenuParser = null;
    int id = -1;
}
	: start:LCURLY  l:STRING  SEMI
          ( n:INT { id = parseNonnegativeInt(n, "menu item"); } )?  SEMI
          { submenuParser = parser.addSubmenu(start, l.getText(), id); }
          ( menuList[submenuParser] )?  SEMI  RCURLY
	;

value[ValueConsumer consumer]
	: scalarValue[consumer]
	| listValue[consumer]
	;

listValue[ValueConsumer consumer]
	: LBRACKET
          ( scalarValue[consumer] ( COMMA scalarValue[consumer] )* )?
          RBRACKET
	;

scalarValue[ValueConsumer consumer]
	: vbf:FALSE_KEYWORD { consumer.addBooleanValue(vbf, false); }
	| vbt:TRUE_KEYWORD { consumer.addBooleanValue(vbt, true); }
	| vi:INT { consumer.addIntegerValue(vi, parseInt(vi.getText())); }
	| vf:FLOAT { consumer.addFloatValue(vf, parseFloat(vf.getText())); }
	| vd:DOUBLE { consumer.addDoubleValue(vd, parseDouble(vd.getText())); }
	| vs:STRING { consumer.addStringValue(vs, vs.getText()); }
        | ve:ENUM_KEYWORD  vei:INT
          { consumer.addEnumerationValue(ve, parseInt(vei.getText())); }
	| compositeValue[consumer]
	;

compositeValue[ValueConsumer consumer]
	: LCURLY  ( float3[consumer] | double3[consumer] )  RCURLY
	;

float3[ValueConsumer consumer]
	: f1:FLOAT  SEMI  f2:FLOAT  SEMI  f3:FLOAT  SEMI
          {
              float x = parseFloat(f1.getText());
              float y = parseFloat(f2.getText());
              float z = parseFloat(f3.getText());
              consumer.addFloat3Value(f1, x, y, z);
          }
	;

double3[ValueConsumer consumer]
	: d1:DOUBLE  SEMI  d2:DOUBLE  SEMI  d3:DOUBLE  SEMI
          {
              double x = parseDouble(d1.getText());
              double y = parseDouble(d2.getText());
              double z = parseDouble(d3.getText());
              consumer.addDouble3Value(d1, x, y, z);
          }
	;

booleanValue returns [boolean value]
	: FALSE_KEYWORD { value = false; }
	| TRUE_KEYWORD  { value = true; }
	;

dumpTokens
	: ( tok:. { System.out.println(tok.toString()); } )*
        ;

//==========================================================================
class GraphFileLexer extends Lexer;
options
{
   k = 2;
   testLiterals = false;
   exportVocab = GraphFile;
   charVocabulary = '\3'..'\377';
   defaultErrorHandler = false;
}

{
    public void tab()
    {
	// NOTE: The number returned by getColumn() is 1-based.
        int column = getColumn() - 1;
        int tabStop = TAB_SIZE * (column + TAB_SIZE) / TAB_SIZE + 1;
        setColumn(tabStop);
    }

    private void syntaxError(String s)
	throws RecognitionException
    {
	String msg = "line " + getLine() + ": column " + getColumn()
	    + ": syntax error: " + s;
	throw new RecognitionException(msg, getFilename(), getLine());
    }

    private static final int  TAB_SIZE = 8;
}

CODE
options { paraphrase = "code string"; }
	: "||"!
	  ( options { greedy = false; }
          : ESCAPE
          | '\n'
            { syntaxError("encountered unescaped newline in code string"); }
	  | ~( '\\' | '\n' )
          )* "||"!
	;

STRING
options { paraphrase = "a string"; }
	: '"'!
          ( ESCAPE
          | '\n' { syntaxError("encountered unescaped newline in string"); }
          | ~( '"' | '\\' | '\n')
          )* '"'!
	;

protected
ESCAPE
	: '\\'
	  ( '\\' { $setText("\\"); }
          | '"' { $setText("\""); }
          | '|' { $setText("|"); }
	  | 'n' { $setText("\n"); }
	  | 'r' { $setText("\r"); }
	  | 't' { $setText("\t"); }
	  | 'f' { $setText("\f"); }
	  | 'b' { $setText("\b"); }
          )
        ;
	exception
	catch [NoViableAltForCharException e]
	{
	    syntaxError("invalid escape sequence '\\" + (char)LA(1) + "'");
	}

KEYWORD
options { testLiterals=true; }
	: ID
	;

IDENT
options { paraphrase = "an identifier"; }
	: '$'! ( WS )? ID
	;
	exception
	catch [RecognitionException e]
        {
	    syntaxError("malformed identifier: "
                        + "need letter or underscore after $");
        }

NUMBER
	: ( '-' )? ( DIGIT )+
	  ( ( '.' ( DIGIT )+ ( EXP )?
              ( 'f'! { $setType(FLOAT); }
              | /* nothing */ { $setType(DOUBLE); }
	      ) 
            )
	  | /* nothing */ { $setType(INT); }
          )
	;
	exception
	catch [RecognitionException e]
        {
            syntaxError("malformed number");
        }

protected
EXP
	: ( 'e' | 'E' ) ( '+' | '-' )? ( DIGIT )+
	;

protected
ID
	: LETTER ( LETTER | DIGIT )*
	;

protected
LETTER
	: ( '_' | 'A'..'Z' | 'a'..'z' )
	;

protected
DIGIT
	: '0'..'9'
	;

SHARP_COMMENT
	: '#' ( ~'\n' )* '\n' { newline(); $setType(Token.SKIP); }
	;

AT_COMMENT
	: '@' ( WS )? ID ( WS )? '=' { $setType(Token.SKIP); }
	;
	exception
        catch [RecognitionException e]
        {
            syntaxError("malformed @-comment: malformed identifier "
	                + "or missing trailing '='");
        }

WHITESPACE
	: WS { $setType(Token.SKIP); }
	;

protected
WS
	: ( ' ' | '\t' | '\f' | '\r' | '\n' { newline(); } )+
	;

LCURLY
options { paraphrase = "'{'"; }
	: '{';

RCURLY
options { paraphrase = "'}'"; }
	: '}';

LBRACKET
options { paraphrase = "'['"; }
	: '[';

RBRACKET
options { paraphrase = "']'"; }
	: ']';

SEMI
options { paraphrase = "';'"; }
	: ';';

COMMA
options { paraphrase = "','"; }
	: ',';
